from typing import NamedTuple, Optional, Set
import pandas as pd
from requests import Session
import json
from functools import lru_cache
from collections import namedtuple

Tag = namedtuple('Tag', ['name', 'ontology_id', 'entity_type'])

class Predictor:
    """
    Class for interacting with curation models
    """
    def __init__(self, token: str) -> None:
        """
        Args:
            token (str): polly user token
        """
        self.session = Session()
        self.session.headers = {
                "Content-Type": "application/vnd.api+json",
                "Cookie": f"refreshToken={token}",
                }
        self.base_url = "https://api.discover.polly.elucidata.io/curations/inferences/"
    
    def run_model(self, identifier: str, input_data: dict) -> dict:
        """
        This is a wrapper around model inference APIs
        It serializes the input data, calls the API and returns deserialized output

        Args:
            identifier (str): A string identifier for the model
            input_data (dict): Model input

        Returns:
            dict

        """
        url = self.base_url + identifier
        payload = json.dumps({
            "data": {
                "attributes": input_data,
                "type": "curation"
              }
        })
        response = self.session.post(url, data=payload)

        try:
            response = response.json()
        except json.JSONDecodeError as e:
            print(response.text)
            raise e
        
        if 'data' in response: 
            return response['data']
        
        return response

    def get_tags(self, text: str, remove_non_specific: bool = False) -> Set[NamedTuple]:
        """
        Returns a set of "tags" inferred from the given piece of text. A tag is just an ontology term.
        This function is a wrapper around `_extract_entities` and `normalize_mention`

        Args:
            text (str): Input text

        Kwargs:
            entities (Optional[Set[str]]): All other types of entities except these are filtered out
            remove_non_specific (bool): If this is True, non specific terms are removed from the output

        Returns: TODO
        """
        # TODO: remove non specific

        entities = self._extract_entities(text, normalize=True)
        res = {Tag(e['name'], e['ontology_id'], e['entity_type']) for e in entities if e['name']}
        return res

    def _extract_entities(self, text: str, entities: Optional[Set[str]] = None, normalize: bool = False):
        """Run an NER model on the given text. The returned value is a list of entities along with span information.

        Args:
            text (str): TODO

        Kwargs:
            entities (Optional[Set[str]]): TODO

        Returns: TODO

        """
        #TODO: If text is too long, break it up into chunks small enough for biobert
        # TODO: Why is this happening?
        text = text.replace('\n', '')
        response = self.run_model("biobert", {"text": text})
        try:
            entities = response['entities']
        except KeyError as e:
            print(response)
            raise e
        
        if normalize:
            for entity in entities:
                norm = self.normalize(entity['keyword'], entity['entity_type'], text)
                entity['ontology_id'] = norm['ontology'] + ':' + norm['ontology_id']
                if norm['ontology'] == 'CUI-less':
                    entity['name'] = None
                else:
                    entity['name'] = norm['name']

        return entities

    @lru_cache(maxsize=None)
    def normalize(self, mention: str, entity_type: str, context: Optional[str] = None) -> dict:
        """Map a given mention (keyword) to an ontology term.

        Args:
            mention (str): Mention of an entity e.g. "Cadiac arrythmia"
            entity_type (str): Should be one of ['disease', 'drug', 'cell_type', 'cell_line', 'species', 'gene']

        Kwargs:
            context (Optional[str]): The text where the mention occurs. This is used to resolve abbreviations

        Returns:
            dict

        """
        #TODO: Composite mentions?
        data = {"mention": {"keyword": mention, "entity_type": entity_type}}
        if context:
            data["context"] = context
            
        output = self.run_model("normalizer", data)
        try:
            term = output['term']
        except KeyError as e:
            return {'ontology': 'CUI-less',
                     'ontology_id': None,
                     'name': None,
                     'entity_type': entity_type}
            
        return output['term']

    def assign_control_pert_labels(self, sample_metadata, columns_to_exclude = None) -> pd.DataFrame:
        """Returns the sample metadata dataframe with 2 additional columns
        `is_control` - whether the sample is a control sample
        `control_prob` - the probability that the sample is control

        Args:
            sample_metadata (pd.DataFrame):

        Kwargs:
            columns_to_exclude (Set[str]): Any columns which don't play any role in determining the label, e.g. any arbitrary sample identifier

        Returns:
            pd.DataFrame

        """
        sample_metadata = sample_metadata.copy()
        
        if columns_to_exclude is None:
            columns_to_exclude = []
        
        cols = sample_metadata.columns.difference(columns_to_exclude)
        
        samples = sample_metadata[cols].to_dict('records')
        request_body = { 'samples': samples }
        response = self.run_model('control-pertubation', request_body)

        response = {k:v for k,v in response.items() if k!= 'version' and k != 'testing'}
        output = pd.DataFrame(response)
        sample_metadata['is_control'] = output['is_control'].values
        sample_metadata['control_prob'] = output['control_prob'].values
        return sample_metadata

    def create_cohorts(self, sample_metadata: pd.DataFrame, columns_to_exclude: Set[str] = None, max_unique: int = 10) -> pd.DataFrame:
        """TODO: Docstring for assign_cohorts.

        Args:
            sample_metadata (pd.DataFrame): TODO

        Kwargs:
            columns_to_exclude (Set[str]): TODO
            max_unique (int): TODO

        Returns: TODO

        """
        # TODO
        raise NotImplementedError()
